#include "CL1602.h"

#include <stdio.h>
#include <string.h>
#include <inttypes.h>
#include "Arduino.h"
#include <Wire.h>

// When the display powers up, it is configured as follows:
//
// 1. Display clear
// 2. Function set: 
//    DL = 1; 8-bit interface data 
//    N = 0; 1-line display 
//    F = 0; 5x8 dot character font 
// 3. Display on/off control: 
//    D = 0; Display off 
//    C = 0; Cursor off 
//    B = 0; Blinking off 
// 4. Entry mode set: 
//    I/D = 1; Increment by 1 
//    S = 0; No shift 
//
// Note, however, that resetting the Arduino doesn't reset the LCD, so we
// can't assume that its in that state when a sketch starts (and the
// CL1602 constructor is called).

CL1602::CL1602()
{
  _i2caddr = MCP23017_ADDRESS;

  _displayfunction = LCD_1LINE | LCD_5x8DOTS;
}

void CL1602::begin(uint8_t cols, uint8_t lines, uint8_t dotsize) {
  Wire.begin();

  if (lines > 1) {
    _displayfunction |= LCD_2LINE;
  }
  _numlines = lines;

  setRowOffsets(0x00, 0x40, 0x00 + cols, 0x40 + cols);  

  // for some 1 line displays you can select a 10 pixel high font
  if ((dotsize != LCD_5x8DOTS) && (lines == 1)) {
    _displayfunction |= LCD_5x10DOTS;
  }

  // SEE PAGE 45/46 FOR INITIALIZATION SPECIFICATION!
  // according to datasheet, we need at least 40ms after power rises above 2.7V
  // before sending commands. Arduino can turn on way before 4.5V so we'll wait 50
  delayMicroseconds(50000);

  // Now we pull both RS and R/W low to begin commands
  _gpioa = 0x00 ;

  writeGPIO(MCP23017_OLATA, _gpioa);
  writeGPIO(MCP23017_OLATB, 0x00);
  writeGPIO(MCP23017_GPPUB, 0x7F); // Pullup switch
  writeGPIO(MCP23017_IODIRA, 0x00); // OUTPUT
  writeGPIO(MCP23017_IODIRB, 0x7F);

  //put the LCD into 4 bit mode
  {
    // this is according to the hitachi HD44780 datasheet
    // figure 24, pg 46

    // we start in 8bit mode, try to set 4 bit mode
    write4bits(0x03);
    delayMicroseconds(4500); // wait min 4.1ms

    // second try
    write4bits(0x03);
    delayMicroseconds(4500); // wait min 4.1ms
    
    // third go!
    write4bits(0x03); 
    delayMicroseconds(150);

    // finally, set to 4-bit interface
    write4bits(0x02); 
  }

  // finally, set # lines, font size, etc.
  command(LCD_FUNCTIONSET | _displayfunction);  

  // turn the display on with no cursor or blinking default
  _displaycontrol = LCD_DISPLAYON | LCD_CURSOROFF | LCD_BLINKOFF;  
  display();

  // clear it off
  clear();

  _gpioa |= LCD_PIN_BL ;

  // Initialize to default text direction (for romance languages)
  _displaymode = LCD_ENTRYLEFT | LCD_ENTRYSHIFTDECREMENT;
  // set the entry mode
  command(LCD_ENTRYMODESET | _displaymode);
}

void CL1602::setRowOffsets(int row0, int row1, int row2, int row3)
{
  _row_offsets[0] = row0;
  _row_offsets[1] = row1;
  _row_offsets[2] = row2;
  _row_offsets[3] = row3;
}

/********** high level commands, for the user! */
void CL1602::clear()
{
  command(LCD_CLEARDISPLAY);  // clear display, set cursor position to zero
  delayMicroseconds(2000);  // this command takes a long time!
}

void CL1602::home()
{
  command(LCD_RETURNHOME);  // set cursor position to zero
  delayMicroseconds(2000);  // this command takes a long time!
}

void CL1602::setCursor(uint8_t col, uint8_t row)
{
  const size_t max_lines = sizeof(_row_offsets) / sizeof(*_row_offsets);
  if ( row >= max_lines ) {
    row = max_lines - 1;    // we count rows starting w/0
  }
  if ( row >= _numlines ) {
    row = _numlines - 1;    // we count rows starting w/0
  }
  
  command(LCD_SETDDRAMADDR | (col + _row_offsets[row]));
}

// Turn the display on/off (quickly)
void CL1602::noDisplay() {
  _displaycontrol &= ~LCD_DISPLAYON;
  command(LCD_DISPLAYCONTROL | _displaycontrol);
}

void CL1602::display() {
  _displaycontrol |= LCD_DISPLAYON;
  command(LCD_DISPLAYCONTROL | _displaycontrol);
}

// Turns the underline cursor on/off
void CL1602::noCursor() {
  _displaycontrol &= ~LCD_CURSORON;
  command(LCD_DISPLAYCONTROL | _displaycontrol);
}

void CL1602::cursor() {
  _displaycontrol |= LCD_CURSORON;
  command(LCD_DISPLAYCONTROL | _displaycontrol);
}

// Turn on and off the blinking cursor
void CL1602::noBlink() {
  _displaycontrol &= ~LCD_BLINKON;
  command(LCD_DISPLAYCONTROL | _displaycontrol);
}

void CL1602::blink() {
  _displaycontrol |= LCD_BLINKON;
  command(LCD_DISPLAYCONTROL | _displaycontrol);
}

// These commands scroll the display without changing the RAM
void CL1602::scrollDisplayLeft(void) {
  command(LCD_CURSORSHIFT | LCD_DISPLAYMOVE | LCD_MOVELEFT);
}

void CL1602::scrollDisplayRight(void) {
  command(LCD_CURSORSHIFT | LCD_DISPLAYMOVE | LCD_MOVERIGHT);
}

// This is for text that flows Left to Right
void CL1602::leftToRight(void) {
  _displaymode |= LCD_ENTRYLEFT;
  command(LCD_ENTRYMODESET | _displaymode);
}

// This is for text that flows Right to Left
void CL1602::rightToLeft(void) {
  _displaymode &= ~LCD_ENTRYLEFT;
  command(LCD_ENTRYMODESET | _displaymode);
}

// This will 'right justify' text from the cursor
void CL1602::autoscroll(void) {
  _displaymode |= LCD_ENTRYSHIFTINCREMENT;
  command(LCD_ENTRYMODESET | _displaymode);
}

// This will 'left justify' text from the cursor
void CL1602::noAutoscroll(void) {
  _displaymode &= ~LCD_ENTRYSHIFTINCREMENT;
  command(LCD_ENTRYMODESET | _displaymode);
}

// Allows us to fill the first 8 CGRAM locations
// with custom characters
void CL1602::createChar(uint8_t location, uint8_t charmap[]) {
  location &= 0x7; // we only have 8 locations 0-7
  command(LCD_SETCGRAMADDR | (location << 3));
  for (int i=0; i<8; i++) {
    write(charmap[i]);
  }
  home() ;
}

void CL1602::backlight(void) {
  _gpioa |= LCD_PIN_BL ;
  writeGPIO(MCP23017_OLATA, _gpioa) ;
}

void CL1602::noBacklight(void) {
  _gpioa &= ~LCD_PIN_BL ;
  writeGPIO(MCP23017_OLATA, _gpioa) ;
}

uint8_t CL1602::buttonRead(uint8_t sw) {
  if (sw > 6)
    return 0 ;

  // read the current GPIO
  return ((readGPIO(MCP23017_GPIOB) >> sw) & 0x1) ? 0 : 1;
}

void CL1602::ledWrite(uint8_t v) {
  writeGPIO(MCP23017_OLATB, v ? 0x80 : 0x00);
}

/*********** mid level commands, for sending data/cmds */

inline void CL1602::command(uint8_t value) {
  send(value, LOW);
}

inline size_t CL1602::write(uint8_t value) {
  send(value, HIGH);
  return 1; // assume sucess
}

/************ low level data pushing commands **********/

// write either command or data
void CL1602::send(uint8_t value, uint8_t mode) {
  _gpioa = (_gpioa & ~LCD_PIN_RS) | (mode ? LCD_PIN_RS : 0) ;
//  writeGPIO(MCP23017_GPIOA, _gpioa) ;

  write4bits(value>>4);
  write4bits(value);

  delayMicroseconds(40);   // commands need > 37us to settle
}

void CL1602::write4bits(uint8_t value) {
  _gpioa = ((value << 4) & 0xF0) | (_gpioa & 0x0F) ;

  _gpioa &= ~LCD_PIN_E ;
  writeGPIO(MCP23017_OLATA, _gpioa) ;
//  delayMicroseconds(1);
  _gpioa |= LCD_PIN_E ;
  writeGPIO(MCP23017_OLATA, _gpioa) ;
//  delayMicroseconds(1);
  _gpioa &= ~LCD_PIN_E ;
  writeGPIO(MCP23017_OLATA, _gpioa) ;
}

/**
 * Writes a given register
 */
void CL1602::writeGPIO(uint8_t reg, uint8_t value) {
	Wire.beginTransmission(_i2caddr);
	Wire.write(reg);
	Wire.write(value);
	Wire.endTransmission();
}

/**
 * Reads a given register
 */
uint8_t CL1602::readGPIO(uint8_t reg) {
	Wire.beginTransmission(_i2caddr);
	Wire.write(reg);
	Wire.endTransmission();
	Wire.requestFrom(_i2caddr, (uint8_t)1);
	return Wire.read();
}
